#!/usr/bin/env python3
import pandas as pd
import re
import gzip
import shutil
from pathlib import Path

# Remove unwanted characters from anywhere is file (should only be in fasta headers)
def modify(file_path):
    # Open and read the contents of the file
    with open(file_path, "r") as f_in:
        contents = f_in.read()
        
    # Preform substitutions for unwanted characters
    contents = re.sub(r"[:,/()=#\x27]", "", contents)
    contents = re.sub(r"[: ]", "_", contents)

    # Write the modified contents back to file
    with open(file_path, "w") as f_out:
        f_out.write(contents)
    return

# =============================================================================
# def modify2(file_path):
#     # Open and read the contents of the file
#     with open(file_path, "rb") as f_in:
#         contentsB = f_in.read()
#         
#     # Preform substitutions for unwanted characters
#     table = bytes().maketrans(b" ", b"_")
#     contentsS = contentsB.translate(None, b":,/()=#").translate(table).decode()
# 
#     # Write the modified contents back to file
#     with open(file_path, "w") as f_out:
#         f_out.write(contentsS)
# =============================================================================

# Un gzip the downloaded NCBI genomes
def decompress(file_path):
    # Open the .gz file and decompress it
    with gzip.open(file_path, "rb") as f_in, open(file_path[:-3], "wb") as f_out:
        shutil.copyfileobj(f_in, f_out) # copy
    return

# Rename amplicon fasta headers to origin contig
def amp_replace(outdir, genus, name):
    df_sum = pd.read_csv(f"{outdir}/amplicons/{genus}-{name}.summary", sep = "\t", header = None, names = ["AmpId", "SequenceId", "PositionInSequence", "Length", "Misc"])
    dict_sum = dict(zip(df_sum.AmpId, df_sum.SequenceId))
    with open (f"{outdir}/amplicons/{genus}-{name}.temp.amplicons", "r") as f_in, open(f"{outdir}/amplicons/{genus}-{name}.amplicons", "w") as f_out:
       for line in f_in:
           if ">amp" in line:
               amp = line.strip().strip(">")
               line = line.replace(amp, dict_sum[amp] + f"_{amp.strip('amp_')}")
           f_out.write(line)
    Path.unlink(f"{outdir}/amplicons/{genus}-{name}.temp.amplicons")
    return

def pairwise_to_csv(pairwise_match, gcf_species, outdir, genus, name):
    pairwise_save_df = pd.DataFrame(pairwise_match, index = gcf_species.values())
    pairwise_save_df.to_csv(f"{outdir}/amplicons/{genus}-{name}_confusion.csv", sep = ",", index = False)
    return
    


#!/usr/bin/env python3
"""
To add:
    When a barrnap result file ends up empty(or just no 16S maybe?) take note and warn then user to check the barrnap logs and write to the logs which file had no 16S in it
"""
import multiprocessing
import subprocess
from pathlib import Path
import shlex


# Spawning the shell call
def call_proc_barrnap(infile):
    # Building the command
    command = f"barrnap --kingdom bac --quiet --threads 1 --reject 0.90 -outseq {infile}.rRNA {infile}"
    # Passing the command to shell piping the stdout and stderr
    subprocess.run(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return

# Multithreading the barrnap calls
def barnap_call(outdir, threads):
    with multiprocessing.Pool(threads) as pool: # spawn the pool
        all_fna = [str(i) for i in list(Path(f"{outdir}/genbank/bacteria/").glob('*/*.fna'))] # generate list of files ending in .rna
        pool.map(call_proc_barrnap, all_fna)
    return
        

# Fishes out 16S sequences and saves them to file
def barrnap_process(in_RNA):
    count = 0 # genome wide count: we are not going to use this
    GCF = str(Path(in_RNA).parent).split(r"/")[-1]
    with open(in_RNA, "r") as f_in, open(f"{in_RNA}.16S", "w") as f_out: 
        for line in f_in:
            if line.startswith(">16S_rRNA"):
                f_out.write(">" + GCF + "_" + line.strip().strip(">16S_rRNA::").split(":")[0] + f"_{count}\n") # write the fasta header to file removing >16S_rRNA:: and adding a counter
                f_out.write(next(f_in)) # write next line also
                count += 1 # incriment count by one
    return

# Splits barrnap output sequences from barrnap_process output
def barrnap_split(in_16S):
    dir_16S = (Path(in_16S).parent / "indiv_16S_dir") # directory path for individual 16S files
    dir_16S.mkdir(exist_ok=True, parents=True) # create directory
    with open(in_16S, "r") as f_in: 
        for line in f_in:
            if line.startswith(">"):
                count = line.strip()[-1]
                with open(f"{dir_16S}/{count}.fna", "w") as f_out:
                    f_out.write(line) # write the fasta header to file removing >16S_rRNA:: and adding a counter
                    f_out.write(next(f_in)) # write next line also
    return

# Concatinates fished out 16S barrnap output into a single file
def barrnap_conc(genus, outdir):
    all_16S = [str(i) for i in list(Path(f"{outdir}/genbank/bacteria/").glob('*/*.16S'))]
    full_path = Path(f"{outdir}/full")
    full_path.mkdir(parents = True, exist_ok = True)
    with open(f"{full_path}/{genus}.16S", "w") as f_out:
        for file in all_16S:
            with open(file, "r") as f_in:
                f_out.write(f_in.read())
    return
        
    


#!/usr/bin/env python3

import subprocess
import shlex
from pathlib import Path
import os

def vsearch_call(outdir, genus, name, ident, log_dir, threads):
    # Defining in and out files/dirs
    infile = f"{outdir}/amplicons/{genus}-{name}.amplicons"
    outfile = f"{outdir}/amplicons/{genus}-{name}.uc"
    clusdir = f"{outdir}/amplicons/{name}-clusters/{genus}-{name}-clus"
    
    Path.mkdir(Path(clusdir).parent, exist_ok = True, parents = True) # making outdir
    
    # Building the command
    command = f"vsearch --cluster_fast {infile} --id {ident} --strand both --uc {outfile} --clusters {clusdir} --quiet --threads {threads}"
    with open(f"{log_dir}/vsearch_{name}.out", "w") as out, open(f"{log_dir}/vsearch_{name}.err", "w") as err:
        # Passing the command to shell piping the stdout and stderr
        subprocess.run(shlex.split(command), stdout = out, stderr = err)
    
    if os.stat(f"{log_dir}/vsearch_{name}.err").st_size != 0:
        raise Exception(f"Something went wrong with VSEARCH\n\nPlease check: {log_dir}/vsearch_{name}.err")
    return



#!/usr/bin/env python3

import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import scipy
from matplotlib.backends.backend_pdf import PdfPages

def heatmap_meta(gcf_species):
    # Turn gcf species cross dictionary into series
    species_series = pd.Series(gcf_species)
    
    # Generate colour palet from this
    palette  = sns.color_palette("husl", len(species_series.unique()))
    species_palette = dict(zip(species_series.unique(), palette))
    row_palette = species_series.map(species_palette)
    return row_palette, species_series

def cluster_heatmap(cluster_dict, row_palette, species_series):
    # Turn cluster dictionary into a dataframe and transpose
    cluster_df = pd.DataFrame.from_dict(cluster_dict).transpose()
    
    # Generate clustering on binary data
    row_clus = scipy.cluster.hierarchy.linkage(np.where(cluster_df > 0, 1, 0), method = "ward")
    col_clus = scipy.cluster.hierarchy.linkage(np.where(cluster_df.transpose() > 0, 1, 0), method = "ward")
    
    # Clustering heatmap
    plot_clus = sns.clustermap(cluster_df, standard_scale = None, 
                   row_linkage = row_clus, 
                   col_linkage = col_clus,
                   yticklabels = species_series,
                   xticklabels = 1,
                   row_colors = row_palette)
    

    plot_clus.ax_heatmap.tick_params(axis='x', labelsize=8)
    # Change plot configeration #
    box_heatmap = plot_clus.ax_heatmap.get_position() # Get heatmaps position
    # Move row colours to left side
    ax_row_colors = plot_clus.ax_row_colors # get the axis
    box_cols = ax_row_colors.get_position() # recover its position
    ax_row_colors.set_position([box_heatmap.max[0], box_cols.y0, box_cols.width*2, box_cols.height]) # plot new position
    
    # Move dendogram to the left a bit
    ax_row_dendogram = plot_clus.ax_row_dendrogram
    box_dendo = ax_row_dendogram.get_position()
    ax_row_dendogram.set_position([box_dendo.x0+0.026, box_dendo.y0, box_dendo.width, box_dendo.height])
    
    return plot_clus, cluster_df

def pairwise_heatmap(pairwise_match, row_palette, species_series):
    # Turn pairwise dictionary into dataframe
    pairwise_df = pd.DataFrame(pairwise_match, index = pairwise_match.keys())
    
    # Clustering heatmap
    plot_dendo = sns.clustermap(pairwise_df, standard_scale = None,
                   row_colors = row_palette,
                   yticklabels = species_series,
                   xticklabels = 1,
                   method = "ward")
    
    plot_dendo.ax_heatmap.tick_params(axis='x', labelsize=8)
    # Change plot configeration #
    box_heatmap = plot_dendo.ax_heatmap.get_position() # Get heatmaps position
    
    # Move row colours to left side
    ax_row_colors = plot_dendo.ax_row_colors # get the axis
    box_cols = ax_row_colors.get_position() # recover its position
    ax_row_colors.set_position([box_heatmap.max[0], box_cols.y0, box_cols.width*2, box_cols.height]) # plot new position
    
    # Move dendogram to the left a bit
    ax_row_dendogram = plot_dendo.ax_row_dendrogram
    box_dendo = ax_row_dendogram.get_position()
    ax_row_dendogram.set_position([box_dendo.x0+0.026, box_dendo.y0, box_dendo.width, box_dendo.height])
    
    return plot_dendo, pairwise_df

def pdf_save(plot_clus, plot_dendo, outdir, genus, name):
    plots = [plot_clus, plot_dendo]
    with PdfPages(f"{outdir}/amplicons/{genus}-{name}.pdf") as pdf_pages:
        for plot in plots:
            pdf_pages.savefig(plot.fig)
    return
            





#!/usr/bin/env python3

from pathlib import Path
import multiprocessing
from glob import glob
from itertools import repeat

import numpy as np
import pandas as pd

from Bio import SeqIO
from collections import Counter

from Bio import Phylo
import matplotlib
import matplotlib.pyplot as plt

def writer(outdir, genus, stats):
    with open(f"{outdir}/{genus}_summary.tsv", "a") as f_app:
        f_app.write("\t".join(stats) + "\n")

def plot_tree(tree, pdf_out):
    plt.figure(figsize=(25, 35), dpi=200)
    Phylo.draw(tree, do_show = False)
    plt.savefig(pdf_out)
    return

def shannon_calc(alignment_path):
            # Read in alignment file and generate a dist of sequences
            seq_dict = {rec.id : rec.seq for rec in SeqIO.parse(alignment_path, "fasta")}
            n_seqs = len(seq_dict)
            seq_len = len(next(iter(seq_dict.values())))
            
            divs = [-1] * seq_len # empty list to record local shannon diversity on each nucleotide position
            # for each nucleotide position generate local shannon diversity
            for i in range(0,seq_len):
                pos = list(value[i] for value in seq_dict.values()) # get all values of a given position
                # Replicating the function of table() in R using a Counter
                c = Counter(pos)
                probs = np.array(list(c[x] for x in ["A", "G", "C", "T", "-"])) / n_seqs # getting count for each base in a numpy array
                divs[i] = -sum(probs * np.log(probs, where = probs != 0)) # log the probabilities where they are not 0 then sum them
            
            return sum(divs)

def summary_16S_run(in_aln, outdir, genus, fast_mode):
    
    indir = Path(in_aln).parent
    # Paths for all needed files
    mismatch_path   = f"{indir}/ani/ANIm_similarity_errors.tab"
    alignment_path = in_aln
    tree_path = in_aln.replace(".16sAln", ".16sTree")
    pdf_out = in_aln.replace(".16sAln", "_16S_div.pdf")
    
    # Getting sequences name details from first record of the alignment fasta
    with open(alignment_path, "r") as f_in:
        count_16S = 0
        for value, line in enumerate(f_in):
            if value == 0:
                splitname = line.strip().split("_")
            elif line.startswith(">"):
                count_16S += 1
        
        
    
    # Taking fasta header and getting important bits out
    GCF = "_".join(splitname[:2]).strip(">")
    #NZ = splitname[2] # Currently unused?
    genera = splitname[3]
    species = splitname[4]
    
    # If ANI was run then do following
    if Path(mismatch_path).is_file():
        
        # Read in ANI simmilatiry errors (the number of unaligned or non-identical bases)
        mismatch = pd.read_table(mismatch_path, sep="\t", index_col = 0)
        
        if len(mismatch) > 1:
            # Get upper triange of this dataframe
            indices  = np.triu_indices(mismatch.shape[0], k=1) # gets indices of upper triangle k=1 ofsets by 1 to avoid the diagonal center 
            upper_mismatch = mismatch.values[indices] # gets the values of these indicies as a n array
            
            # Calculate stats on this array
            mean_mis = str(round(np.mean(upper_mismatch), 2))
            sd_mis = str(round(np.std(upper_mismatch, ddof = 1), 2))
            max_mis = str(max(upper_mismatch))
            min_mis = str(min(upper_mismatch))
            
            total_div = str(shannon_calc(alignment_path))
            
            # roll_means_30 = np.convolve(divs, np.ones(30), "valid")/30   
            
            if fast_mode:
                pass
            else:
                tree = Phylo.read(tree_path, "newick")
                plot_tree(tree, pdf_out)
            
            stats = [GCF, genera, species, str(count_16S),  mean_mis, sd_mis, max_mis, min_mis, total_div]
            writer(outdir, genus, stats)

        else:
            mean_mis, sd_mis, max_mis, min_mis, total_div = (str(0), str(0), str(0), str(0), str(0))
            stats = [GCF, genera, species, str(count_16S), mean_mis, sd_mis, max_mis, min_mis, total_div]
            writer(outdir, genus, stats)
    else:
        mean_mis, sd_mis, max_mis, min_mis, total_div = ("-", "-", "-", "-", "-")
        stats = [GCF, genera, species, str(count_16S), mean_mis, sd_mis, max_mis, min_mis, total_div]
        writer(outdir, genus, stats)
        
            
def multiproc_sumamry(outdir, genus, threads, fast_mode):
    with multiprocessing.Pool(threads) as pool:
        all_aln = [str(i) for i in list(Path(f"{outdir}/genbank/bacteria/").glob('*/*.16sAln'))]
        pool.starmap(summary_16S_run, zip(all_aln, repeat(outdir), repeat(genus), repeat(fast_mode)))
    return

        






#!/usr/bin/env python3
"""
A program to analyse the usefulness of amplicon sequences

Created under GPL-3.0 license

@author: Robert Murphy
@email: robert.murphy@bio.ku.dk
"""
# Importing requierd libraries
import argparse
import os
from pathlib import Path
import shutil
import multiprocessing
from itertools import repeat




import ngd_download
import barrnap_run
import pcr_run
import pyani_run
import utils
import msa_run
import summary_16S
import vsearch_run
import overlaps
import make_heatmap



# Defining user arguments
def parse_args():
    parser = argparse.ArgumentParser(
        description="RibDif2 evaluates the differences in user defined amplicons  within a genus or species to indicate if that amplicon can differentiate between the taxanomic groups")

    parser.add_argument("-g", dest="genus", 
                        help="The genus you want to search within. E.g. 'Staphylococcus' OR 'Staphylococcus aurea' if wanting to use a species", 
                        required = True)
    
    parser.add_argument("-o", dest = "outdir",
                        help = "Output direcotry path. Default is current directory",
                        default = "False")
    
    # Mutually exclusive group of rerun and clobber
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-r", dest = "rerun",
                        help = "Rerun on same of different primers. Avoids having to redownload the same genomes",
                        action = "store_true") # Action store true will default to false when argument is not present
    
    group.add_argument("-c", dest = "clobber",
                        help="Delete previous run if present in output directory", 
                        action = "store_true") 
    
    parser.add_argument("-p", dest = "primers", 
                        help = "Path to custom primer-file, must be a tab seperated file with name, forward and reverse primers. See default.primers",
                        default = "False")
    
    parser.add_argument("-a", dest = "ANI", 
                        help = "ANI is off by default, turn on if you care about individual genomes. The $genus-summary.csv-file will only contain a list of genomes when off",
                        action = "store_true")
    
    parser.add_argument("-f", dest="frag", 
                        help = "Allow use of fragmented genomes. Full genomes are recomended/requierd for detecting all 16S-genes, use with caution. Off by default",
                        action = "store_true")
    
    parser.add_argument("-m", dest = "msa",
                        help = "make multiple sequence alignment and trees",
                        action = "store_true")
    
    parser.add_argument("-i", dest = "id",
                        help = "Identity to cluster amplicons at if not using the deafult 100%%. e.g. .99. Does not cluster at the genome level, so beware",
                        default = 1)
    
    parser.add_argument("-t", dest = "threads",
                        help = "Number of threads to use. Default is all avaliable",
                        default = os.cpu_count())
    
    parser.add_argument("-s", dest = "speed", 
                        help = "Run in fast mode skipping all but the nessacary steps (No barrnap, msa, ani etc.)",
                        action = "store_true")
    return parser.parse_args()

def arg_handling(args, workingDir):
    print("#= Parsing arguments =#\n\n")
    
    if args.speed:
        print("Running in fast mode skipping all but the nessacary steps (No barrnap, msa, ani etc.)")
    
    #Checking if user is running on a species
    if " " in args.genus: # checking is there is a space in genus/species name
        print(f"Detected species {args.genus.split(' ')[1]}.\n\n")
        genus = args.genus.replace(" ", "_") # if so replacing it with an '_'
    else:
        genus = args.genus


    # Parsing the primers argument
    if args.primers == "False":
        primer_path = workingDir.parent
        primer_file = Path(f"{primer_path}/docs/default.primers")
    else:
        primer_file = args.primers
    # Check primer file exists and is not empty
    if Path(f"{primer_file}").is_file() and os.stat(f"{primer_file}").st_size == 0:
        raise Exception(f"Error: The provided primer file is empty.\n{primer_file}\nPlease provide a populated primer file")

    elif Path(f"{primer_file}").is_file() == False: # If it does not exist then raise this exception
        raise Exception(f"Error: {primer_file} does not exist")
    

    # Checking if user provided own outdir and if not setting to RibDif directory
    if args.outdir == "False":
        outdir = Path(f"{workingDir.parent}/results/{genus}")
    else:
        outdir = Path(args.outdir)
    
    #Resolving rerun argument
    if args.rerun == True:
        if Path(f"{outdir}").is_dir() == False:
            print(f"No records of {genus} exists. Ignoring rerun request and downloading genomes")
            rerun = False
        else:
            print(f"Reusing previously downloaded {genus} genomes")
            rerun = True
    else:
        rerun = False
    
    # Resolving clobber argument
    if args.clobber == True:
        print(f"Removing old run of {genus}" )
        try:
            shutil.rmtree(Path(f"{outdir}")) # Remove genus and all subdirectories
        except FileNotFoundError: # catch if directory not found
            print(f"{genus} folder does not exist, ignoring clobber request")
            pass
    elif Path(f"{outdir}").is_dir() and args.rerun == False: # catch if genus output already exists and rerun was not specified and clobber was not used
       raise FileExistsError(f"""/n/n{outdir} folder already exists. Run again with -c/--clobber, -r/--rerun or set another output directory/n/n""")
       
    print("\n\n#= All arguments resolved =#\n\n")
    
    return genus, primer_file, outdir, rerun


def main():
    workingDir = Path(os.path.realpath(os.path.dirname(__file__))) # getting the path the script is running from
    
    args = parse_args()
    
    print(f"""\n
          #=========================================#
          #== RibDif2 is running on: {args.genus} ==#
          #=========================================#\n\n""")
    
    #Argument handeling
    genus, primer_file, outdir, rerun = arg_handling(args, workingDir)
    
    log_dir = Path(outdir) / "ribdif_logs"
    Path(log_dir).mkdir(exist_ok = True, parents = True)
    
    # If rerun is false, download and handle genomes from NCBI
    if rerun == False:
        # Download genomes from NCBI
        genome_count = ngd_download.genome_download(genus = genus, outdir = outdir, threads = args.threads, frag = args.frag)
    
        # Un gziping fasta files
        with multiprocessing.Pool(args.threads) as pool: # Create a multiprocessing pool with #threads workers
            all_gz = [str(i) for i in list(Path(f"{outdir}/genbank/bacteria/").glob('**/*.gz'))]# Recursively search the directory for .gz files and convert path to string sotring in a list
            pool.map(utils.decompress, all_gz)
        
        # Remove unwanted characters from anywhere is file (should only be in fasta headers)
        print("\n\nModifying fasta headers.\n\n")
        with multiprocessing.Pool(args.threads) as pool:
            all_fna = [str(i) for i in list(Path(f"{outdir}/genbank/bacteria/").glob('*/*.fna'))]
            pool.map(utils.modify, all_fna)
    else:
        genome_count = len(list(Path(f"{outdir}/genbank/bacteria").glob("*")))
        print(f"\n\n{genome_count} previously downloaded genomes of {genus} were found")
            
    # If using default primers call barrnap and rerun is false - this assume
    if args.primers == "False" and not args.speed: # Is skipping barrnap actually faster?
        print("#= Running barrnap on downloaded sequences =#\n\n")
        barrnap_run.barnap_call(outdir, threads = args.threads)
        
        # Processing barrnap output > fishing out 16S sequences
        with multiprocessing.Pool(args.threads) as pool:
            all_RNA = [str(i) for i in list(Path(f"{outdir}/genbank/bacteria/").glob('*/*.rRNA'))]
            #gene_num = [*range(len(all_RNA))] # adding gene num count here (in congruence with v1. Could also just use in silico pcr amp count)
            pool.map(barrnap_run.barrnap_process, all_RNA) # removed zip(gene_num) and was originally starmap
        
        # Concatinate all 16S to one file
        barrnap_run.barrnap_conc(genus, outdir)
        
        #If ANI is true calculate
        if args.ANI:
            # First need to split whole 16S sequences into seperate files
            with multiprocessing.Pool(args.threads) as pool:
                all_16S = [str(i) for i in list(Path(f"{outdir}/genbank/bacteria/").glob('*/*.16S'))]
                pool.map(barrnap_run.barrnap_split, all_16S)
               
            # Call pyani
            print("Calculating intra-genomic mismatches and ANI for each genome.\n\n")
            pyani_run.pyani_call(outdir, args.threads)
        else:
            print("Skipping detailed intra-genomic analysis and ANI (if needed, use -a/--ANI).\n\n")
        
        if args.speed:
            pass
        else:
            # ALignment of full 16S genes recoverd from barrnap
            print("Alligning full-length 16S genes within genomes with muscle and building trees with fastree.\n\n")
            msa_run.muscle_call_multi(outdir, args.threads)
        
        # Genome statistic summary
        with open(f"{outdir}/{genus}_summary.tsv", "w") as f_out:
            f_out.write("GCF\tGenus\tSpecies\t#16S\tMean\tSD\tMin\tMax\tTotalDiv\n")
        summary_16S.multiproc_sumamry(outdir, genus, args.threads, args.speed)
        
        # Running msa on concatinated 16S sequences
        if args.msa == True:
            print(f"Alligning all {genus} 16S rRNA genes with muscle and building tree with fasttree.\n")
            infile , outAln, outTree = f"{outdir}/full/{genus}.16S", f"{outdir}/full/{genus}.16sAln", f"{outdir}/full/{genus}.16sAln" # Asigning in and out files
            msa_run.muscle_call_single(infile, outAln, outTree)
        else:
            print("Skipping alignments and tree generation for 16S rRNA genes (if needed, use -m/--msa).\n\n")
            
        # PCR for default primers
        infile = f"{outdir}/full/{genus}.16S" # path to concatinated 16S barrnap output
        names =  pcr_run.pcr_call(infile, outdir, genus, primer_file, workingDir)
        

        

    # PCR for custom primers   
    else:
        # Concatinate all downloaded genomes
        all_fna = [str(i) for i in list(Path(f"{outdir}/genbank/bacteria/").glob('*/*.fna'))]
        with open(f"{outdir}/genbank/bacteria/{genus}_total.fna", "w") as f_out:
            for file in all_fna:
                with open(file, "r") as f_in:
                    f_out.write(f_in.read())
                    
        
        infile = f"{outdir}/genbank/bacteria/{genus}_total.fna" # path to cocatinate genus genomes
        names = pcr_run.pcr_call(infile, outdir, genus, primer_file, workingDir)
        #pcr_run.pcr_parallel_call(outdir, genus, primer_file, workingDir, threads)
        #pcr_run.pcr_cleaner(outdir, primer_file, genus)
    
    for name in names:
        # Rename amplicon fasta headers to origin contig
        utils.amp_replace(outdir, genus, name)
        
        
    # msa on all amplicons
    if args.msa == True:
        print("Alligning all amplicons with Muscle and building tree with fasttree.\n")
        for name in names:
            infile , outAln, outTree = f"{outdir}/amplicons/{genus}-{name}.amplicons", f"{outdir}/amplicons/{genus}-{name}.aln", f"{outdir}/amplicons/{genus}-{name}.tree" # Asigning in and out files
            msa_run.muscle_call_single(infile, outAln, outTree)
    else:
        print("Skipping alignments and trees generation for amplicons (if needed, use -m/--msa).\n\n")
    
    print ("Making unique clusters with vsearch.\n\n")
    for name in names:
        vsearch_run.vsearch_call(outdir, genus, name, args.id, log_dir, args.threads)
    
    if args.msa == True:
        print("Making amplicon summary file for tree viewer import.\n\n")
        msa_run.format_trees(outdir, genus, name)
    
    # Generating the heatmaps #
    for name in names:
        print(f"Making heatmaps for {name}\n\n")
        # Cleaning vsearch clustering data
        all_gcfs, uc_dict_clean, gcf_species, cluster_count = overlaps.uc_cleaner(outdir, genus, name)
        
        # Generate a dictionary (that will become a matrix) of GCF cluster membership  
        cluster_dict = overlaps.cluster_matrix(all_gcfs, uc_dict_clean, cluster_count)
        
        # Find all species overlap in the cluster_dict
        combinations = overlaps.species_overlap(cluster_dict, cluster_count, gcf_species)
        
        # Find all GCF overlaps in the cluster dictionary
        pairwise_match = overlaps.gcf_overlaps(all_gcfs, uc_dict_clean, gcf_species)
        utils.pairwise_to_csv(pairwise_match, gcf_species, outdir, genus, name)
        
        # Generate metadata for heatmaps
        row_palette, species_series = make_heatmap.heatmap_meta(gcf_species)
        
        # Plot the cluster matrix
        plot_clus, cluster_df = make_heatmap.cluster_heatmap(cluster_dict, row_palette, species_series)
        
        # Plot the GCF overlap matrix
        plot_dendo, pairwise_df = make_heatmap.pairwise_heatmap(pairwise_match, row_palette, species_series)
        
        # Save the heatmaps
        make_heatmap.pdf_save(plot_clus, plot_dendo, outdir, genus, name)
        
        overlaps.overlap_report(combinations, gcf_species, cluster_df, genus, name)

if __name__ == '__main__':
    main()


#!/usr/bin/env python3
import ncbi_genome_download as ngd
from pathlib import Path

# Using Kai's NCBI genome downloader to get all genones of the specified genus
# Avaliable at: https://github.com/kblin/ncbi-genome-download
def genome_download(genus, outdir, threads, frag):
    
    if frag: # if the user wants fragmented genomes
        print(f"Downloading all genome recoreds of {genus} fron NCBI")
        ngd.download(section='genbank', 
               file_formats = 'fasta', 
               genera = genus,  
               output = outdir, 
               parallel = threads*2,
               groups = 'bacteria')
        
        
    else: # if the user (correctly) only want complete genomes
        print(f"Downloading all complete genome records of {genus} from NCBI")
        ngd.download(section='genbank', 
               file_formats = 'fasta', 
               genera = genus,  
               output = outdir, 
               assembly_levels = "complete",
               parallel = threads*2,
               groups = 'bacteria')
    
    if Path(f"{outdir}/genbank/bacteria").is_dir():
        count = len(list(Path(f"{outdir}/genbank/bacteria").glob("*")))
        print(f"\n\n{count} genomes of {genus} were downloaded")
        return count
    else:
        raise Exception(f"Download failed because {genus} is invalid or there are no records of the requested type in NCBI\n\n")
        return


#!/usr/bin/env python3

import pandas as pd
import numpy as np



# Import and clean the cluster file
def uc_cleaner(outdir, genus, name):
    
    uc_path = f"{outdir}/amplicons/{genus}-{name}.uc"
    
    # read in cluster file
    uc_df = pd.read_csv(uc_path, sep = "\t", header = None)
    # remove cluster summaries
    uc_df_clean = uc_df.loc[uc_df[0] != "C"].copy() # Explicitly setting as a copy to avoid SettingWithCopyWarning warnings
    
    # Create new columns
    uc_df_clean.loc[:, "GCF"] = ["_".join(i.split("_")[:2]) for i in uc_df_clean[8]] # GCF column
    uc_df_clean.loc[:, "Species"] = [i.split("_")[4] for i in uc_df_clean[8]]  # Species column
    
    # Sorting dataframe
    #uc_df_clean = uc_df_clean.sort_values(by = ["GCF"]) # check if this matters later on
    uc_df_clean = uc_df_clean.sort_values(by = ["Species", "GCF"]) # Mikael orderd by both? why?
    
    # Moving unclassified species to the bottom
    if 'sp.' in uc_df_clean.Species.unique():
        uc_df_clean = pd.concat([uc_df_clean[uc_df_clean.Species != "sp."], uc_df_clean[uc_df_clean.Species == "sp."]])
    
    # Storing information needed later
    uc_dict_clean = uc_df_clean.T.to_dict("list") # Chaging the dataframe to a dictionary
    gcf_species = dict(zip(uc_df_clean.GCF, uc_df_clean.Species)) # dicttionary of GCF to species
    all_gcfs = uc_df_clean.GCF.unique() # getting an array of all GCFs
    cluster_count = len(uc_df_clean[1].unique()) # getting a count of how many cluster there are
    return all_gcfs, uc_dict_clean, gcf_species, cluster_count
    

    
# Generate a dictionary (that will become a matrix) of GCF cluster membership  
def cluster_matrix(all_gcfs, uc_dict_clean, cluster_count):
    # Creating a dictionary of lists where keys are GCFs and values are lists the length of cluster present
    cluster_dict = {key: [0]*cluster_count for key in all_gcfs}
    
    # Looping over unique GCFs
    for gcf in all_gcfs:
        # Fishing out current GCFs and converting to a dictionary of lists  where rows indicies in the df are keys
        current_GCF = dict([(k, v) for k, v in uc_dict_clean.items() if v[10] == gcf])
        #current_GCF = uc_df_clean[uc_df_clean.GCF == gcf].T.to_dict("list") # if using dataframe
        # Loop through fished out GCFs clusters
        for r in current_GCF.values():
            # For the current cluster for a given GCF add 1 to the count of time the given GCF is present in the given cluster
            cluster_dict[r[10]][r[1]] = cluster_dict[r[10]][r[1]] + 1
    return cluster_dict



# Find all species overlap in the cluster_dict
def species_overlap(cluster_dict, cluster_count, gcf_species):
    combinations = []
    # Loop through all clusters
    for i in range(cluster_count):
        # Get all GCF values that have more that one amplicon beloning to the current cluster
        keys = [k for k, v in cluster_dict.items() if v[i] > 1]
        # Get unique species names asigned for the GCFs recoverd above
        cluster_species = set([gcf_species[gcf] for gcf in keys])
        # If more than one species belongs to this cluster then save that combination
        if len(cluster_species) > 1: 
            combinations.append("/".join(cluster_species))
    return combinations


# Find all GCF overlaps in the cluster dictionary
def gcf_overlaps(all_gcfs, uc_dict_clean, gcf_species):
    # Dictionary of lists of length equal to all gcfs populated with list length of all gcfs
    pairwise_match = {key: [0]*len(all_gcfs) for key in all_gcfs}
    for gcf in all_gcfs:
        # Fish out current GCF
        current_GCF = dict([(k, v) for k, v in uc_dict_clean.items() if v[10] == gcf])
        # current_GCF = uc_df_clean[uc_df_clean.GCF == gcf].T.to_dict("list") # if using dataframe
        # get species of current GCF
        #species = gcf_species[gcf]
        # get unique cluster this GCF belongs to
        clusters = set([v[1] for v in current_GCF.values()])
        # Getting a list of other GCFs that are members of the clusters the current GCF belongs to
        clusMatchGCF = set([v[10] for v in uc_dict_clean.values() if v[1] in clusters])
        
        for m in clusMatchGCF:
            gcf_index = np.where(all_gcfs==m)[0][0]
            pairwise_match[gcf][gcf_index] = 1
    return pairwise_match


def overlap_report(combinations, gcf_species, cluster_df, genus, name):
    
    genome_count = len(cluster_df.index)
    named_genomes = len([s for s in gcf_species.values() if s != "sp."]) # Recovering species names that are not equal so "sp."
    unamed_genomes = len(gcf_species) - named_genomes # Subtracting names species from total specpes
    unq_species = len([i for i in set(gcf_species.values()) if i != "sp."]) # unique names speices
    # turning the cluster_df binary with np.where and summing rows then counting the times rowsum is greater that 1
    multi_allele = sum(np.where(cluster_df > 0, 1, 0).sum(axis = 1) > 1)
    if combinations: # are the unique entries into combinations greater than 0. Not sure we need the set.
        unq_combs = sorted(set(combinations), key = len, reverse = True) # getting unique combinations and sorting by reverse string length
        has_overlap = set([e for c in combinations for e in c.split("/")]) # getting a unique set of species that experienced overlap
        if "sp." in has_overlap:
            has_overlap.remove("sp.")
    else:
        unq_combs = []
        has_overlap = []
        
    count_overlap = len(has_overlap)
    with open("saveme.txt", "r+") as f_out:
        f_out.write(f"""Summary of {genus} differentiation by {name} amplicons:\n
                    Genomes downloaded: {genome_count}
                    \tWith species name: {named_genomes}
                    \tWithout species name: {unamed_genomes}
                    \tUnique species names: {unq_species}\n
                    {multi_allele} of {genome_count} ({round(100*multi_allele/genome_count, 2)}%) genomes have multiple alleles.
                    {count_overlap} of {unq_species} ({round(100*count_overlap/unq_species, 2)}%) species have at least one overlap.\n\n""")
        if unq_combs:
            for i in range(len(unq_combs)):
                members = unq_combs[i].split("/")
                f_out.write(f"Group {i}:\t" + ' / '.join(members) + "\n")
        f_out.seek(0)
        print(f_out.read())


#!/usr/bin/env python3
from glob import glob
import multiprocessing
import subprocess
import shlex
from pathlib import Path

# Spawning the shell call
def call_proc_pyani(indir):
    aniDir = Path(indir).parent / "ani"
    # Building the command
    command = f"average_nucleotide_identity.py -i {indir} -o {aniDir}"
    # Passing the command to shell piping the stdout and stderr
    subprocess.run(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return


# Multithreading the pyani calls
def pyani_call(outdir, threads):
    with multiprocessing.Pool(threads) as pool: # spawn the pool
        all_16S_dirs = [i for i in glob(f"{outdir}/genbank/bacteria/*/indiv_16S_dir")] # can we use a faster method than glob?
        pool.map(call_proc_pyani, all_16S_dirs)
    return




#!/usr/bin/env python3
#import multiprocessing
import subprocess
from pathlib import Path
#from itertools import repeat
import shlex
"""
Implement a producer and consumer setup for writing the pcr output whe multiprocessing: https://stackoverflow.com/questions/11196367/processing-single-file-from-multiple-processes

I removed the counter from all functions for labeling output files as it seems the cocatinated fasta ran pretty fast.
"""

# Spawning the shell call
def call_proc_pcr(infile, outdir, genus, name, fwd, rvs, workingDir): # removed , counter
    # Building the command
    command = f"perl {workingDir}/in_silico_PCR.pl -s {infile} -a {fwd} -b {rvs} -r -m -i > {outdir}/amplicons/{genus}-{name}.summary 2> {outdir}/amplicons/{genus}-{name}.temp.amplicons"
    # Passing the command to shell piping the stdout and stderr
    with open(f"{outdir}/amplicons/{genus}-{name}.summary", "w") as f_std, open(f"{outdir}/amplicons/{genus}-{name}.temp.amplicons", "w") as f_err:
        subprocess.run(shlex.split(command), stdout = f_std, stderr = f_err)
    return 

# Multithreading the in silico pcr calls
# =============================================================================
# def pcr_parallel_call(outdir, genus, primer_file, workingDir, threads):
#     amplicon_dir = Path(f"{outdir}/amplicons")
#     amplicon_dir.mkdir(parents = True, exist_ok = True)
#     print("\n\nGenerating amplicon sequences")
#     with open(primer_file, "r") as f_primer:
#         for primer in f_primer:
#             name, fwd, rvs = primer.split("\t")
#             with multiprocessing.Pool(threads) as pool: # spawn the pool
#                 all_fna = [str(i) for i in list(Path(f"{outdir}/genbank/bacteria/").rglob('*.fna'))] # generate list of files ending in .fna
#                 counter = range(len(all_fna))
#                 pool.starmap(call_proc_pcr, zip(all_fna, repeat(outdir), repeat(genus), repeat(name), repeat(fwd), repeat(rvs), repeat(workingDir), counter))
#     return
# =============================================================================

def pcr_call(infile, outdir, genus, primer_file, workingDir):
    amplicon_dir = Path(f"{outdir}/amplicons")
    amplicon_dir.mkdir(parents = True, exist_ok = True)
    print("#= Generating amplicon sequences =#\n\n")
    with open(primer_file, "r") as f_primer:
        names = []
        for primer in f_primer:
            name, fwd, rvs = primer.strip().split("\t")
            call_proc_pcr(infile, outdir, genus, name, fwd, rvs, workingDir)
            names.append(name)
    return names

# Cleaning multithreaded PCR call (concatinating etc)
# =============================================================================
# def pcr_cleaner(outdir, primer_file, genus):
#     print("cleaning files")
#     amplicon_dir = Path(f"{outdir}/amplicons")
#     with open(primer_file, "r") as f_primer:
#         for primer in f_primer:
#             name = primer.split("\t")[0]
#             print(f"looking for {name}")
#             
#             # Summary file
#             summary_file = f"{amplicon_dir}/{genus}-{name}.summary"
#             print(f"master summary: {summary_file}")
#             all_sum = [str(i) for i in list(Path(f"{amplicon_dir}/").rglob(f"*{name}.summary*"))] # getting all summary files
#             for file in all_sum: # looping over them
#                 with open(file, "r") as f_in: # open each one
#                     if Path(summary_file).is_file(): # if the master summary file already exists
#                         with open(f"{summary_file}", "a") as f_sum: # open it in append mode
#                             lines = f_in.read().splitlines()
#                             f_sum.write("\n".join(lines[1:])) # and write everything from the second line
#                     else:
#                         with open(f"{summary_file}", "w") as f_sum:
#                             f_sum.write(f_in.read())
#             # Amplicon file
# 
# =============================================================================


#!/usr/bin/env python3

import multiprocessing
from pathlib import Path
import subprocess
import shlex
from Bio import Phylo
import pandas as pd
import re


# Spawning the shell call
def call_proc_muscle(infile):
    outfile = str(Path(infile).parent / Path(infile).stem)
    # Building the command
    command1 = f"muscle -align {infile} -output {outfile}.16sAln" # Do I need to strip the alignement file of white space and commas?
    command2 = f"fasttree -quiet -nopr -gtr -nt {outfile}.16sAln"
    # Passing the command to shell piping the stdout and stderr
    subprocess.run(shlex.split(command1), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    with open(f"{outfile}.16sTree", "w") as f_std:
        subprocess.run(shlex.split(command2), stdout = f_std, stderr = subprocess.PIPE)
    
    return


# Multithreading the pyani calls
def muscle_call_multi(outdir, threads):
    with multiprocessing.Pool(threads) as pool: # spawn the pool
        all_16S = [str(i) for i in list(Path(f"{outdir}/genbank/bacteria/").glob('*/*.16S'))]
        pool.map(call_proc_muscle, all_16S)
    return



# Calling Muscle for MSA of all 16S sequences
def muscle_call_single(infile, outAln, outTree):
    # Building the command
    command1 = f"muscle -align {infile} -output {outAln}" # Do I need to strip the alignement file of white space and commas? # Do we want to use -distance1 kbit20_3
    command2 = f"fasttree -quiet -nopr -gtr -nt {outAln}"
    subprocess.run(shlex.split(command1), stdout = subprocess.PIPE, stderr = subprocess.PIPE)
    with open(outTree, "w") as f_std:
        subprocess.run(shlex.split(command2), stdout = f_std, stderr = subprocess.PIPE)
    return

def format_trees(outdir, genus, name):
    uc_path = f"{outdir}/amplicons/{genus}-{name}.uc"
    tree_path = f"{outdir}/amplicons/{genus}-{name}.tree"
    out_path = f"{outdir}/amplicons/{genus}-{name}-meta.tsv"
    
    # read in cluster file
    uc_df = pd.read_csv(uc_path, sep = "\t", header = None)
    uc_df_clean = uc_df.loc[uc_df[0] != "C"]
    
    # read in tree and get lead labels
    tree = Phylo.read(tree_path, "newick")
    tree_terminals = tree.get_terminals() # getting all leaf info
    tips = [t.name for t in tree_terminals] # getting just leaf names
    tips_clean = [re.sub(r"_chromosome_.*|_complete_.*|_genome_.*", "", n) for n in tips] # cleaning leaf names

    
    # Getting individual name parts with list comprehension
    GCF = ["_".join(i.split("_")[:2]) for i in tips_clean]
    NZ = [i.split("_")[2] for i in tips_clean]
    species = [i.split("_")[4] for i in tips_clean]
    strain = ["_".join(i.split("_")[5:])for i in tips_clean]
    n_gene = [i.split("_")[-1] for i in tips]
    
    # Getting cluster number for each item of tips in the order of tips # Why not just reorder tips to that of the dataframe?
    uc_df = uc_df.sort_values(by = tips)
    match_index =[uc_df_clean.index[uc_df_clean[8] == i] for i in tips]
    clusters = [uc_df.loc[m][1].values[0] for m in match_index]
    
    meta_df = pd.DataFrame({"name": tips, "GCF":GCF, "NZ":NZ, "Species":species, "Strain":strain, "nGene":n_gene, "Cluster":clusters})
    meta_df.to_csv(out_path, sep = "\t", index = False)
    
  



# -*- coding: utf-8 -*-
"""RibDif2 - Ribosomal differentiation

Documentation: https://github.com/Rob-murphys/RibDif2

RibDif2 contains the following moduels:
    
    Ribdif
"""
__authors__ = "Robert Murphy"
__licence__ = "GPL-3.0"
__version__ = (1, 1, 1)


# Try on python v3.5
import sys as _sys
if _sys.version_info[:2] < (3, 7):
    raise ImportError('Python version must be >= 3.7')